require "chipmunk"

require_relative "physic_consts"

class Numeric
  def radians_to_vec2
    CP::Vec2.new(Math::cos(self), Math::sin(self))
  end
end

class PhysicEngine
  MARQUEE = 10

  def initialize
    super
    @cp = CP::Space.new
    @cp.damping = DAMPING
    @classes = []
    #for deleting shapes of deleted bodies
    @shapes = {}
    @constraints = {}
    puts "Using Chipmunk version #{CP::VERSION}"
  end

  def register(aclass)
    @classes << aclass
  end

  def end_registration
    #define groups for classes
    @groups = Hash.new(0)
    ngroup = 1
    cls = @classes + [Walls]
    cls.each do |c|
      if c.method_defined? ("dontcollide_#{c.name.downcase}".to_sym)
        @groups[c] = ngroup
        puts "physics group #{ngroup}: #{c}"
        ngroup += 1
      end
    end

    #define layers:
=begin
    @layers = Hash.new(0)
    nlayer = 0
    cls.each do |c1|
      cls.each do |c2|
        next if c1.to_s > c2.to_s
        if c1.method_defined?("collide_#{c2.name.downcase}".to_sym) ||
            c2.method_defined?("collide_#{c1.name.downcase}".to_sym)
          @layers[c1] += (1<<nlayer)
          @layers[c2] += (1<<nlayer)
          puts "physics layer #{nlayer}: #{c1} with #{c2}"
          nlayer += 1
          throw :too_much_layers if nlayer >= 31
        end
      end
    end
=end

    @classes.each do |c1|
      @classes.each do |c2|
        if c1.method_defined? ("collide_#{c2.name.downcase}".to_sym)
          puts "Registering collision: #{c1} with #{c2}"
          @cp.add_collision_func(c1, c2) do |shape1, shape2|
            obj1 = shape1.object
            obj2 = shape2.object
            #unless... is preventing possible crash
            obj1.send("collide_#{c2.name.downcase}".to_sym, obj2) unless shape1.body.p == shape2.body.p
          end
        elsif c1.method_defined? ("dontcollide_#{c2.name.downcase}".to_sym)
          puts "Registering no collision: #{c1} with #{c2}"
          @cp.add_collision_func(c1, c2, &nil)
        end
      end
    end
  end

  def addbody(owner, material)
    body = CP::Body.new(material[:mass], material[:inertia])
    @cp.add_body(body)
    body.object = owner
    body
  end

  def add_static_body(owner)
    body = CP::Body.new_static
    #@cp.add_body(body)
    body.object = owner
    body
  end

  def add_circle_shape(body, radius, x, y)
    shape = CP::Shape::Circle.new(body, radius, CP::Vec2.new(x, y))
    shape.collision_type = body.object.class
    shape.object = body.object
    @shapes[body] = [] unless @shapes[body]
    @shapes[body] << shape
    @cp.add_shape(shape)
    shape
  end

  def fit_shape(shape, aclass, material)
    shape.group = @groups[aclass]
    #shape.layers = @layers[aclass]
    shape.e = material[:elast]
    shape.u = material[:friction]
  end

  def count_shapes
    @shapes.values.inject(0) { |sum, x| sum += x.length }
  end

  def add_poly_shape(body, shape_array)
    throw vertexes.inspect unless CP::Shape::Poly.valid? shape_array
    shape = CP::Shape::Poly.new(body, shape_array, CP::ZERO_VEC_2)
    shape.collision_type = body.object.class
    shape.object = body.object
    @shapes[body] = [] unless @shapes[body]
    @shapes[body] << shape
    @cp.add_shape(shape)
    shape
  end

  def add_line_shape(body, x1, y1, x2, y2)
    if MARQUEE > 0
      #      x1 -= MARQUEE*(x2<=>x1)
      #      y1 -= MARQUEE*(y2<=>y1)
      #      x2 += MARQUEE*(x2<=>x1)
      #      y2 += MARQUEE*(y2<=>y1)
    end
    shape = CP::Shape::Segment.new(body, CP::Vec2.new(x1, y1), CP::Vec2.new(x2, y2), MARQUEE)
    shape.object = body.object
    shape.collision_type = body.object.class
    @shapes[body] = [] unless @shapes[body]
    @shapes[body] << shape
    @cp.add_shape(shape)
    shape
  end

  def del_body(body)
    clear_body body
    @shapes.delete body
    @cp.remove_body(body)
  end

  def del_shape(shape)
    @cp.remove_shape(shape)
    @shapes.each_value do |sha|
      sha.delete shape
    end
  end

  def add_constraint(cons, owner)
    @cp.add_constraint cons
    @constraints[owner] = [] unless @constraints[owner]
    @constraints[owner] << cons
  end

  def remove_constraint(cons, owner)
    @cp.remove_constraint cons
    @constraints[owner].delete cons
  end

  def clear_body(body)
    if @shapes[body]
      @shapes[body].each do |x|
        @cp.remove_shape x
      end
    end
    @shapes[body] = []
    if @constraints[body]
      @constraints[body].each do |x|
        @cp.remove_constraint x
      end
    end
    @constraints[body] = []
  end

  def update
    SUBSTEPS.times { @cp.step(PHYS_DT) }
  end
end

# Convenience method for converting from radians to a Vec2 vector.
class Numeric
  def radians_to_vec2
    CP::Vec2.new(Math::cos(self), Math::sin(self))
  end
end
